#/***************************************************************************
# *   Copyright (C) 2008 by Paul Hardwick                                   *
# *   paul@peck.org.uk                                                      *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU General Public License as published by  *
# *   the Free Software Foundation; either version 2 of the License, or     *
# *   (at your option) any later version.                                   *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU General Public License for more details.                          *
# *                                                                         *
# *   You should have received a copy of the GNU General Public License     *
# *   along with this program; if not, write to the                         *
# *   Free Software Foundation, Inc.,                                       *
# *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
# ***************************************************************************/
_rev = "$Revision: 16 $"[10:-1].strip()
__version__ = "1.2."+_rev
__author__ = "Paul Hardwick <paul@peck.org.uk>"
__date__ = "$Date: 2009-05-26 23:04:11 +0100 (Tue, 26 May 2009) $"[6:18].strip()

#HSOserial
import time
import re
import os, os.path, sys
if os.environ.get("HSOCDEBUG",False):
    dmode = True
else:
    dmode = False
    
try:
    import serial #use latest py-serial module if installed
except:
    import hsocserial as serial #fall back version

resms = re.compile('''"SM",(\d+,\d+)''')
resmsallhead = re.compile('''\+CMGL: (\d+),\d*,.*,\d+''')
resmsghead = re.compile('''\+CMGR: \d+,.*,\d+''')
recdshead = re.compile('''\+CDS: (\d+)''')
remsgref = re.compile('''\+CMGS: (\d+)''')
reohcip = re.compile('''_OHCIP: (\d+)''') #HSDPA call in progress status
resmspdu = re.compile('([0-9ABCDEF]+)')
resys = re.compile("_OSSYSI: (\d)")
reowcti= re.compile("_OWCTI: (\d)") #WCMA cell type _OWCTI: 0
reocti= re.compile("_OCTI: \d,(\d)") #GSM cell type _OCTI: 1,2
reowancall=re.compile("_OWANCALL: \d,(\d)") #IP connection state
resig = re.compile("_OSIGQ: (\d+),\d")
recreg = re.compile("\+CREG: (\d)")
recgreg = re.compile("\+CGREG: (\d)")
recgregcommand = re.compile("\+CGREG: (\d),(\d)")
recops = re.compile('''\+COPS: \d,\d,"(.+)",(\d)''')
reoldcops = re.compile('''\+COPS: \d,\d,"(.+)"''')
reopsys = re.compile('''_OPSYS: (\d),(\d)''')
recsq = re.compile("\+CSQ: (\d+),\d")
relistcops = re.compile('''(\d,".+?",".+?","\d+?",\d)''')
reoldlistcops = re.compile('''(\d,".+?",".+?","\d+?")''')
reopsel=re.compile('''(\d),"(.+?)","(.+?)","(\d+?"),(\d)''')

recmti = re.compile('''\+CMTI: "..",\s*(\d+)''')
recbm = re.compile('''\+CBMI: "..",(\d+)''')
recms = re.compile('''\+CMS ERROR: (\d+)''')
recsca = re.compile('''\+CSCA: "(\+?\d+)",(\d+)''')
altrecsca = re.compile('''\+CSCA: "([0-9a-fA-F]+)",(\d+)''')
redata = re.compile('''\((\d*\.\d*.*?)\).*\((\d*\.\d*.*?)\)''')

class HSOserial:
    "The serial class using the HSO driver"
    def cprint(self, *args):
        for n in args:
            print n

    def _sendreceive(self, command, timeout=5, report=True):
        echoOn=True
        "A rountine to send an AT command then return answering strings"
        if timeout:
            oldtime = self.ser.timeout
            try:
                self.ser.setTimeout(timeout)
            except:
                print "Device IO error:", command.strip()
                if report: self.ErrorDialog('error', 'error_comms')
                return (0, '')
        if True: #
        #try:
            #self.ser.write(command+"\r\n")
            self.ser.write(command+"\r")
            if echoOn:
                for n in range(10):
                    if True:#self.ser.inWaiting():
                        res = self.ser.readline().strip()
                        self.cprint( "\n", `res`)
                        ret = ()
                        if res == command:
                            break
                else:
                    self.cprint( "DUff response", `res`)
                    ret = (0, res)

                if not ret:
                    res = self.ser.readline()
                    self.cprint("#", `res`)

                while len(res) and len(res.strip()) == 0:
                    res = self.ser.readline()
                    self.cprint( "@", `res`)
                ret = (1, res.strip())
                if ret[1] != "OK" and ret[1] != "ERROR":
                    res = self.ser.readline()
                    self.cprint("%", `res`)
                    res = self.ser.readline()
                    self.cprint("$", `res`)
            if timeout:
                self.ser.setTimeout(oldtime)
            return ret
        else: #
        #except:
            if report: self.ErrorDialog('error', 'error_comms')
            return (0, '')
        
    def __init__(self, port, getpin, prog, cprint=None, ):
        "Serial class init"
        self.command_stack = []
        if cprint !=None:
            #self.cprint =cprint
            pass
        self.waitcnt = 0 #count of _OBLS loop
        self.pinrequired = "False"
        self.port = port
        self.mode = prog['umts_mode']
        self.initstrings = [prog.get("initstring1", ''), prog.get("initstring2", ''), prog.get("initstring3", '')]
        self.connectproc=False
        self.prog = prog
        self.ser = None

    def pause(self, period):
        "A pause generator"
        now = time.time()
        wait = now+period
        while time.time() < wait:
            time.sleep(0.1)

            
    def isOpen(self):
        "Is the port open and active"
        if self.ser:
            return self.ser._isOpen
        return False
        
    def dialogs(self, warn, error):
        "links to the GUI"
        self.WarningDialog = warn
        self.ErrorDialog = error
    
    def initserial(self):
        "Initiaise the serial port"
        self.cprint("Init ser")
        if os.path.exists(self.port):
            ret = self.openport(self.port, self.initstrings, self.cprint)
            if ret:
                self.ser = ret
                ret = self._sendreceive('ATE1')
                return self.ser
        return None

    def setStatusCallbacks(self, sig, operator, registration, 
                           tech, connected, disconnected, received, sent, 
                           showNetworks, 
                           antennaState):
        "Link up Serial device to the main program"
        self.signal = sig
        self.operator = operator
        self.registration = registration
        self.tech = tech
        self.connected = connected
        self.disconnected = disconnected
        self.received_data = received
        self.sent_data = sent
        self.showNetworks=showNetworks
        self.antennaState = antennaState
        
    def checkpin(self):
        "Check PIN requirements"
        ret = self._sendreceive('AT+CLCK="SC",2')
        if ret[0] == 1:
            if ret[1].find("ERROR") != -1: #Something wrong with SIM
                return "ERROR"
            elif ret[1]=="+CLCK: 0":
                pass          
            elif ret[1]=="+CLCK: 1":
                pass
            else: 
                #an error in the SIM access
                return None
        elif ret[0] ==0:
            return "ERROR"
        ret = self._sendreceive("AT+cpin?")
        if ret[1].find("+CPIN: ")==0:
            pinr = ret[1][7:]
            if pinr in ['PH-NET PIN','PH-SIMLOCK PIN', 'PH-CORP PUK', 'PH-CORP PIN', 'PH-SP PUK',
                    'PH-SP PIN', 'PH-NETSUB PUK', 'PH-NETSUB PIN', 'PH-NET PUK', 'PH-NET PIN']:
                return pinr #Some network lock feature
            if pinr in ['READY', 'SIM PIN2']:
                return "READY" #No PIN required
            elif pinr in ['SIM PUK', 'SIM PUK2']:
                return pinr # SIM PUK PUK2 required
            elif pinr == "SIM PIN":
                #SIM PIN required
                return pinr
        return None

    def scanNetworks(self):
        "Send the scan networks command"
        self.command('AT+COPS=?')

    def setNetwork(self, op=None):
        "Manual Setting of the network"
        if op:
            self.command('AT+COPS=1,2,"%s"'%op)
            #self.command('''AT+COPS=3,0''') #make sure we continue to report in th right mode ;)
            self.requestStatus()
            
    def secondaryLED(self, state):
        if state: self.command("AT_OSWLED=1")
        else: self.command("AT_OSWLED=0")
            
    def requestStatus(self):
        "Trigger report on signal a registration"
        self.command("AT+CSQ")
        self.command("AT+CGREG?")    
            
    def autoNetwork(self):
        "Set auto network mode"
        self.command('''AT+COPS=3,0''') #just in case a modem fogets what mode it is in ;)
        self.command('AT+COPS=0')
        self.requestStatus()

    def sendpin(self, pin):
        "Send the PIN to the SIM"
        ret = self._sendreceive('AT+CPIN="%s"'%pin)
        if ret[1] =="+CME ERROR: incorrect password":
            return False
        else:
            return True

    def waitforsubsystem(self):
        "Hang around until device sub-systems are ready"
        ret = self._sendreceive('''AT_OBLS''')
        self.waitcnt = (self.waitcnt+1)%32
        self.signal(self.waitcnt)
        if (ret[1].find("_OBLS: 1,1,1")==0) :
            return True
        else:
            return False

    def initAT(self):
        "Set up the AT interface and unsolicited messaging"
        self.waitforsubsystem()
        self.ser.write("AT_OWANCALL=1,0,1\r")
        time.sleep(0.5)
        self.ser.flush()
        ret = self._sendreceive("ATE1")
        self.cprint(ret)      
        ret = self._sendreceive("ATI")
        self.cprint(ret)           
        if ret[0]:
            self.ATI = ret[1]  
        ln = self.ser.readlines()
        ret = self._sendreceive("AT_OPMN")
        self.cprint(ret)      
        if ret[1].find("Model: GI0031") != -1 :
            self.device ="2G"
        else:
            self.device = "3G"
        if self.device == "3G" :
            ret = self._sendreceive('''AT_OCTI=1''') #enable 2G cell type unsol' msgs
            self.cprint(ret)        
            ret = self._sendreceive('''AT_OSSYS=1''')
            self.cprint(ret)   
            ret = self._sendreceive('''AT_OPSYS=%s,2'''%(self.mode))
            self.cprint(ret)
        self.isUMTS = '0'#set mode to '2g' initially
        ret = self._sendreceive('''AT_OSQI=1''')
        self.cprint(ret)
        ret = self._sendreceive('''AT+CNMI=2,1,0,1,0''')
        self.cprint(ret)  
        ret = self._sendreceive('''AT+CREG=1''')
        self.cprint(ret)
        ret = self._sendreceive('''AT+CGREG=1''')
        self.cprint(ret)
        ret = self._sendreceive('''AT+COPS=3,0''')
        self.cprint(ret)
        ret = self._sendreceive('''AT_OSWLED=0''') #if supported - switch off LED
        self.cprint(ret)
        ret = self._sendreceive('''AT_OANTPOS=1''') #if supported - switch on antenna position indicator
        self.cprint(ret)

        self.ATI = "Not defined"

        self.CGMI = "Not defined"
        ret = self._sendreceive("AT+CGMI")
        self.cprint(ret)
        if ret[0]:
            self.CGMI = ret[1]
        self.GMR = "Not defined"
        ret = self._sendreceive("AT+GMR")
        self.cprint(ret)
        if ret[0]:
            self.GMR = ret[1]
        self.GSN = "Not defined"
        ret = self._sendreceive("AT+CGSN")
        self.cprint(ret)
        if ret[0]:
            self.GSN = ret[1]
        
        return (self.CGMI, self.GMR, self.GSN)

    def setMode(self, mode):
        "Set 3G mode"
        self.mode=mode
        self.command('AT_OPSYS=%s,2'%self.mode)
        self.command("AT_OPSYS?")

    def closeport(self):
        "Close serial port"
        global command_stack
        command_stack = []
        for n in range(10, 0, -1):
            if not self.status(): break

        try:
            ret = self._sendreceive('''AT+CMGF=0''', timeout=1, report=False)
            self.cprint(ret)
            ret = self._sendreceive('''AT_OSQI=0''', timeout=1, report=False)
            self.cprint(ret)
            ret = self._sendreceive('''AT+CNMI=0''', timeout=1, report=False)
            self.cprint(ret)
            ret = self._sendreceive('''AT+CREG=0''', timeout=1, report=False)
            self.cprint(ret)
            ret = self._sendreceive('''AT+CGREG=0''', timeout=1, report=False)
            self.cprint(ret)
            ret = self._sendreceive('''AT_OSSYS=0''', timeout=1, report=False)
            self.cprint(ret)
            self.ser.write("AT_OWANCALL=1,0,1\r")
            time.sleep(0.2)
            self.ser.flush()
        except:
            pass
        try:
            self.ser.close()
        except:
            print "Could not close port"
            return       
        print "Closed port"

    def command(self, cmd):
        "push AT commands a sequences on to stack for processing by periodic"
        self.command_stack.append(cmd)

    def pre_status(self):
        "Pre periodic AT commands to seed GUI"
        self.command("AT+csq")
        self.command("AT+cgreg?")
        self.command("AT+COPS?")
        self.command("AT_OPSYS?")
        #self.command("AT+csq")
        
 

    def checkUnMessages(self, ret):
        "service unscheduled messages in the receive queue"
        if ret.find("OK")==0:
            return True
        if ret.find("AT")==0:
            return True
        if ret.find("ERROR")==0:
            return True       
        if ret.find("_OSIGQ: ")==0:
            sig = resig.search(ret)
            if sig:
                sig = sig.groups()[0]
                if sig:
                    self.signal(int(sig))
            return True
        elif ret.find("_OCTI:")==0: #2G cell type
            typ = reocti.search(ret)
            if typ:
                typ = typ.groups()[0]
                if typ:
                    self.tech(False, typ)
            return True
        elif ret.find("_OWCTI:")==0: #2G cell type
            typ = reowcti.search(ret)
            if typ:
                typ = typ.groups()[0]
                if typ:
                    self.tech(True, typ) 
            return True
        elif ret.find("_OANTPOSI: 1") ==0:
            #antenna out
            self.antennaState(True)
            return True
        elif ret.find("_OANTPOSI: 0") ==0:
            #antenna in
            self.antennaState(False)
            return True
            
        elif ret.find("_OWANDATA: ")==0:
            params = ret.split(',')
            if len(self.prog['connection']['ipaddress']) :
                IP = self.prog['connection']['ipaddress']
            else:
                IP = params[1]                
            if self.prog['connection']['autodns'] == '0':
                DNS1 = self.prog['connection']['dns1']
                DNS2 = self.prog['connection']['dns2']
            else:
                AutoDNS = '1'
                DNS1 = params[3]
                DNS2 = params[4]
            up = "hsolinkcontrol up %s %s %s %s"%(self.prog['procdev'], DNS1, DNS2, IP)
            print up
            result = os.system(up)
            print "hsolinkcontrol response:", result
            if not result:
                self.connectproc = True
                self.connected()
                self.command("AT_OHCIP")
            else:
                self.disconnected()
                self.ErrorDialog('error',self.str['error_hsolinkcontrol']%( up))               
            return True
                            
        elif ret.find("_OWANCALL:")==0:
            cs = reowancall.search(ret)
            if cs:
                cs = cs.groups()[0]
                if cs:
                    if cs =='0':
                        #disconnected
                        print "disconnected"
                        if self.connectproc == True:
                            sts = os.system("hsolinkcontrol down " +self.prog["procdev"])
                            if sts != 0 :
                                self.ErrorDialog('error', 'error_hsolinkcontrol')
                            self.connectproc = False
                        self.cprint("\ndisconnectd")            
                        self.requestStatus()            
                        self.disconnected()
                    elif cs == '1':
                        #connected message not get IP data
                        self.connectproc = False
                        self.connecting()
                    elif cs == '2':
                        #call failed
                        print "Connection failed"
                        pass
            return True
        elif ret.find("_OSSYSI: ")==0:
            sys = resys.search(ret)
            if sys:
                sys = sys.groups()[0]
                if sys:
                    if sys == '2':
                        self.command("AT_OWCTI?")
                        self.tech(True, '0')
                    else:
                        self.command("AT_OCTI?")
                        self.tech(False, '0')                
            return True
        elif ret.find("+CGREG: ")==0:
            sig = recgreg.search(ret)
            if sig:
                sig = sig.groups()[0]
                self.registration(sig)
                if sig in ["1", "5"]:
                    self.command("AT+cops?")
                else:
                    self.operator('-------')
            return True
        elif ret.find("+CREG: ")==0:
            sig = recreg.search(ret)
            if sig:
                sig = sig.groups()[0]
                self.registration(sig)
                if sig in ["1", "5"]:
                    self.command("AT+cops?")
                else:
                    self.operator('-------')
            return True
        return False        
        
    def checkSchMessages(self, ret):
        "service scheduled messages"
        if ret.find("+CSQ: ")==0:
            sig = recsq.search(ret)
            if sig:
                sig = sig.groups()[0]
                if sig:
                    self.signal(int(sig))
            return True
        elif ret.find("_OHCIP:")==0:#HSDPA call status
            typ = reohcip.search(ret)
            if typ:
                typ = typ.groups()[0]
                if typ == '1':
                    self.tech(True, '2')
                return True
        elif ret.find("_OCTI:")==0: #2G cell type
            typ = reocti.search(ret)
            if typ:
                typ = typ.groups()[0]
                if typ:
                    self.tech(False, typ)
            return True
        elif ret.find("_OWCTI:")==0: #3G cell type
            typ = reowcti.search(ret)
            if typ:
                typ = typ.groups()[0]
                if typ:
                    self.tech(True, typ)        
            return True
        elif ret.find("+COPS: ")==0:
            ops = recops.search(ret)
            if ops:
                op = ops.groups()[0]
                self.isUMTS = ops.groups()[1]
                self.operator(op.upper())
                if self.isUMTS in '2456':
                    self.command("AT_OWCTI?")
                    self.tech(True, '0')
                else:
                    self.command("AT_OCTI?")
                    self.tech(False, '0')
                self.operator( op.upper())
            else:
                ops = reoldcops.search(ret)
                try:
                    if ops:
                        op = ops.groups()[0]
                        code = ops.groups()[1]
                        self.operator(op.upper())
                    else:
                        ops = relistcops.findall(ret)
                        if ops:
                            self.command("AT+cgreg?")
                            self.showNetworks(ops)
                        else:
                            ops = reoldlistcops.findall(ret)
                            if ops:
                                self.starting=False
                                ops = [ n+',0' for n in ops]
                                self.command("AT+cgreg?")
                                self.showNetworks(ops)
                except:
                    self.starting=False
            return True
        elif ret.find("_OPSYS: ")==0:
            sys = reopsys.search(ret)
            if sys:
                self.mode = sys.groups()[0]
            return True        
        elif ret.find('OK') != -1:
            return True     
        return False        

    def _pinenable(self):
        pass
    def _pindisable(self):
        pass
    def _pinchange(self):
        pass
    
    def runcommand(self):
        "get special commands off stack"
        cmd = self.command_stack.pop(0)
        if cmd == "!#pinenable":
            self._pinenable()
        elif cmd == "!#pindisable":
            self._pindisable()
        elif cmd == "!#pinchange":
            self._pinchange()
        elif cmd == "!#connect":
            self._connect()
        elif cmd == "!#disconnect":
            self.dis_connect()        
        elif cmd == "!#disconnected":
            self.disconnected()
    
    def status(self):
        "periodically called by HSOconnect"
        if True: #try:
            self.ser.setTimeout(0.1)
        else:#except:
            #during periodic comms the link has failed
            return False
        while True:
            ans = self.ser.readline()
            ret = ans.strip()
            if ans == '':# Nothing in the receive buffer so break out
                break
            if ret == '': #An empty line - get next line
                continue
            self.cprint("<<"+ret+">>\n")
            #Unscheduled msgs
            if not self.checkUnMessages(ret):
                if not self.checkSchMessages(ret):
                    if ret.find('AT') == -1:
                        print "Unhandled msg", `ret`
                        self.cprint("Unprocessed msg", `ret`, '\n')
        #process command stack
        if len(self.command_stack):
            if self.command_stack[0][0:2]=='!#': #special handling command
                self.runcommand()
            else:
                self.cprint("Sending ->", self.command_stack[0], '\n')
                if True: #try:
                    self.ser.write(self.command_stack.pop(0)+"\r")
                else:#except:
                    return False
        return True
#####################################################

    def connect(self):
        "Handle connect button event -"
        if self.connectproc:
            self.dis_connect()
        else: 
            self.justconnect()

            
    def dis_connect(self):
        "Disconnect from the internet"
        if self.connectproc == True:
            os.system("hsolinkcontrol down "+self.prog['procdev'])
            self.connectproc = False
            self.ser.write("AT_OWANCALL=1,0,1\r")
            time.sleep(0.2)
            self.ser.flush()
            self.cprint("\ndisconnectd")            
            self.requestStatus()
            self.command("!#disconnected")
            return True
        else:
            return False
            
 
    def  justconnect(self):
        self.command("!#connect")
        return False

    def _connect(self):
        "do the internet connection"
        if not self.connectproc:
            res = self._sendreceive('AT+CGDCONT=1,"IP","%s"'%self.prog['connection']['apn'])
            res = self._sendreceive('AT$QCPDPP=1,1,"%s","%s"'%(self.prog['connection']['password'], self.prog['connection']['username']))
            if res[0] and (res[1] == "ERROR"):
                res = self._sendreceive('AT_OPDPP=1,1,"%s","%s"'%(self.prog['connection']['password'], self.prog['connection']['username']))
            self.command("AT_OWANCALL=1,1,1")

        
    def connecting(self):
        "called periodically during connection request to see if we have IP details"
        if self.connectproc:
            return False
        self.command("AT_OWANDATA=1")
        return True

    def openport(self, port, initstrings, outfunc):
        "open the serial inteerface"
        port = os.path.realpath(port)
        try:
            outfunc("Port ", port, '\n')
            self.ser = serial.Serial(port, 115200, timeout=1)
            outfunc('Detected\n')
        except serial.SerialException , n:
            outfunc(str(n)+'\n')
            return None
        if  self.ser.isOpen():
            outfunc("Port open\n")
            for t in range(10):
                s =  self.ser.readline()
                if s == '': break
            for n in range( 30):
                ret = self._sendreceive( "AT")
                if ret == None: return -1
                if ret[1] != "OK":
                    time.sleep(0.1)
                    continue
                else: break
            if ret[1] != "OK": return -1 #port open but no communication
            ret = self._sendreceive( "AT+CFUN?") 
            if ret and ret[1]=="+CFUN: 0":
                ret = self._sendreceive("AT+CFUN=1") 
                res =  self.ser.readline()
            if initstrings:
                for n in initstrings:
                    if len(n):
                        outfunc(n, '\n')
                        ret = self._sendreceive(n)
                        outfunc(ret, '\n')
            return  self.ser
        outfunc("not open\n")
        return None
        
        
    def doexit(self, txt1, txt2):
        "Try and exit"
        print "####################Trying to exit ########################"
        try:
            self.ErrorDialog(txt1, txt2)
        except:
            print "Exit on serial error"
            sys.exit(1)
        
if __name__ == "__main__":
    hso = HSOserial('/dev/ttyHS1')
    ret = hso.initserial()
    if ret == 'False':
        hso.initAT()
        print dir(hso)
    
